""":mod:`futureutils` --- Iterators implementing `futures and promises`_
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. codeauthor:: Hong Minhee <minhee@dahlia.kr>

This module provides several trivial wrappers of iterable object that
makes iterators running in parallel easily and simply. Why its name is
:mod:`futureutils` is that it introduces the concept of `futures and promises`_
into Python iterators and generators.

It works well on Python 2.5+ --- CPython 2.5+, PyPy 1.4+.

.. _futures and promises: http://en.wikipedia.org/wiki/Futures_and_promises

"""
import sys
import functools
import threading

__author__ = 'Hong Minhee <minhee@dahlia.kr>'
__license__ = 'MIT License'
__version__ = '0.9'


def promise(iterable):
    """Promises the passed ``iterable`` object and returns its future
    iterator.

    .. sourcecode:: pycon

       >>> import time, datetime
       >>> def myiter():
       ...     for x in xrange(5):
       ...         yield x
       ...         time.sleep(0.5)
       ...
       >>> it = promise(myiter())
       >>> time.sleep(2)
       >>> start = datetime.datetime.now()
       >>> list(it)
       [0, 1, 2, 3, 4]
       >>> delta = datetime.datetime.now() - start
       >>> delta.seconds
       0
       >>> delta.microseconds > 500000
       True

    It could be used for simple parallelization of IO-bound iterable objects.

    It propagates an inner exception during iteration also as well as a normal
    iterator:

    .. sourcecode:: pycon

       >>> def pooriter():
       ...     yield 1
       ...     raise Exception('future error')
       ...
       >>> it = promise(pooriter())
       >>> it.next()
       1
       >>> it.next()
       Traceback (most recent call last):
         ...
       Exception: future error

    :param iterable: an iterable object to promise
    :type iterable: iterable object
    :returns: a promised future iterator
    :rtype: iterable object

    .. seealso:: Decorator :func:`future_generator()`

    """
    result = []
    finished = [False]
    lock = threading.Lock()
    def iterate(iterable, result, finished, lock):
        try:
            for el in iterable:
                lock.acquire()
                result.append(el)
                lock.release()
        except Exception, e:
            finished[0] = e
            finished.append(sys.exc_info()[2])
        else:
            finished[0] = True
    thread = threading.Thread(target=iterate,
                              args=(iterable, result, finished, lock))
    thread.start()
    def iterator():
        while True:
            i = 0
            for i, el in enumerate(result):
                yield el
            lock.acquire()
            del result[:i + 1]
            lock.release()
            if isinstance(finished[0], Exception):
                raise finished[0], None, finished[1]
            elif finished[0]:
                break
            thread.join(0.05)
        thread.join()
    return iterator()


def future_generator(function):
    """The decorator that makes the result of decorated generator ``function``
    to be promised and return a future iterator.

    It's a simple decorator wrapper of :func:`promise()` for generator
    functions.

    .. sourcecode:: pycon

       >>> import time, datetime
       >>> @future_generator
       ... def mygenerator():
       ...     for x in xrange(5):
       ...         yield x
       ...         time.sleep(0.5)
       ...
       >>> it = mygenerator()
       >>> time.sleep(2)
       >>> start = datetime.datetime.now()
       >>> list(it)
       [0, 1, 2, 3, 4]
       >>> delta = datetime.datetime.now() - start
       >>> delta.seconds
       0
       >>> delta.microseconds > 500000
       True

    :param function: a generator function to make to future generator
    :type function: callable object
    :returns: a future generator function
    :rtype: callable object

    .. seealso:: Function :func:`promise()`

    """
    @functools.wraps(function)
    def promised_generator(*args, **kwargs):
        return promise(function(*args, **kwargs))
    return promised_generator


if __name__ == '__main__':
    import doctest
    doctest.testmod()

